;; Generated package description from ztree.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "ztree" "1.0.7.0.20260726.0" "Text mode directory tree" '((cl-lib "0")) :commit "48f17807ac6c765a66cf29b2fcdd08f1aa7da7d6" :authors '(("Alexey Veretennikov" . "alexey.veretennikov@gmail.com")) :maintainer '("Alexey Veretennikov" . "alexey.veretennikov@gmail.com") :keywords '("files" "tools") :url "https://github.com/fourier/ztree")
